import time

from .base import AugerBaseApi
from .trial import AugerTrialApi
from ..exceptions import AugerException

class AugerExperimentSessionApi(AugerBaseApi):
    """Auger Experiment Api."""

    def __init__(self, ctx, experiment_api=None,
        session_name=None, session_id=None):
        super(AugerExperimentSessionApi, self).__init__(
            ctx, experiment_api, session_name, session_id)

    def list(self, params=None):
        params = {} if params is None else params
        params['project_id'] = self.parent_api.parent_api.oid
        return super().list(params)

    def run(self):
        try:
            return self._call_update({'id': self.object_id, 'status': 'preprocess'})
            # return self.rest_api.call(
            #     'update_experiment_session',
            #     {'id': self.object_id, 'status': 'preprocess'})
        except Exception as e:
            self.ctx.log(
                'Start experiment session failed. Try one more time ...: %s' % (e))

            #Try one more time
            time.sleep(60)
            return self._call_update({'id': self.object_id, 'status': 'preprocess'})
            # return self.rest_api.call(
            #     'update_experiment_session',
            #     {'id': self.object_id, 'status': 'preprocess'})

        # self.wait_for_status(['waiting', 'preprocess', 'started'])

    def interrupt(self):
        try:
            status = self.status()
            if status in ['waiting', 'preprocess', 'started']:
                self.rest_api.call(
                    'update_experiment_session',
                    {'id': self.object_id, 'status': 'interrupted'})
                return True
        except Exception as e:
            if 'Event \'interrupted\' cannot transition' not in str(e):
                raise e
        return False

    def create(self):
        evaluation_options, model_type, dataset_statistics = \
            self.parent_api.get_experiment_settings()
        return self._call_create({
            'experiment_id': self.parent_api.object_id,
            'model_settings': evaluation_options,
            'dataset_statistics': dataset_statistics,
            'model_type': model_type})

    def get_leaderboard(self):
        #TODO: create leaderboard same way as in auger-ml
        trial_api = AugerTrialApi(self.ctx, self)
        leaderboard, score_name = [], None
        for item in iter(trial_api.list()):
            score_name = item.get('score_name')
            leaderboard.append({
                'model id': item.get('id'),
                'algorithm': item.get('hyperparameter').\
                    get('algorithm_name').split('.')[-1],
                item.get('score_name'):\
                    '{0:.4f}'.format(item.get('score_value'))
            })

        if score_name:
            leaderboard.sort(key=lambda t: list(t.values())[-1], reverse=False)
            
        return leaderboard



import os

from azureml.core import Dataset
from a2ml.api.utils import fsclient, get_remote_file_info
from .project import AzureProject
from .exceptions import AzureException
from a2ml.api.utils.decorators import error_handler, authenticated
from .credentials import Credentials


class AzureDataset(object):

    def __init__(self, ctx, ws = None):
        super(AzureDataset, self).__init__()
        self.ctx = ctx
        self.ws = ws
        self.credentials = Credentials(self.ctx).load()

    @error_handler
    @authenticated
    def list(self):
        selected = self.ctx.config.get('dataset', None)
        datasets = Dataset.get_all(self._get_ws())
        ndatasts = len(datasets)
        for name in datasets.keys():
            self.ctx.log(
                ('[%s] ' % ('x' if selected == name else ' ')) + name)
        self.ctx.log('%s DataSet(s) listed' % ndatasts)
        return {'datasets': [name for name in datasets.keys()]}

    @error_handler
    @authenticated
    def create(self, source = None, validation=False):
        ws = self._get_ws(True)
        if source is None:
            source = self.ctx.config.get('source', None)
        if source is None:
            raise AzureException('Please specify data source file...')

        if source.startswith("http:") or source.startswith("https:"):
            url_info = get_remote_file_info(source)
            if self.ctx.config.get('source_format', "") == "parquet" or \
               url_info.get('file_ext', "").endswith(".parquet"):
                dataset = Dataset.Tabular.from_parquet_files(path=source)
            else:        
                dataset = Dataset.Tabular.from_delimited_files(path=source)

            dataset_name = url_info.get('file_name')+url_info.get('file_ext')
        else:
            with fsclient.with_s3_downloaded_or_local_file(source) as local_path:
                ds = self.ws.get_default_datastore()

                if self.ctx.config.path and not local_path.startswith("/"):
                    local_path = os.path.join(self.ctx.config.path, local_path)

                ds.upload_files(files=[local_path], relative_root=None,
                    target_path=None, overwrite=True, show_progress=True)
                dataset_name = os.path.basename(local_path)
                if dataset_name.endswith(".parquet") or self.ctx.config.get('source_format', "") == "parquet":
                    dataset = Dataset.Tabular.from_parquet_files(path=ds.path(dataset_name))
                else:    
                    dataset = Dataset.Tabular.from_delimited_files(
                        path=ds.path(dataset_name))

        dataset.register(workspace = ws, name = dataset_name,
            create_new_version = True)
        self._select(dataset_name, validation)
        self.ctx.log('Created DataSet %s' % dataset_name)
        return {'dataset': dataset_name}

    @error_handler
    @authenticated
    def delete(self, name = None):
        ws = self._get_ws()
        if name is None:
            name = self.ctx.config.get('dataset', None)
        if name is None:
            raise AzureException('Please specify dataset name...')
        ds = Dataset.get_by_name(ws, name)
        ds.unregister_all_versions()
        self._select(None, False)
        self.ctx.log('Deleted dataset %s' % name)
        return {'deleted': name}

    @error_handler
    @authenticated
    def select(self, name = None):
        self._select(name)
        self.ctx.log('Selected dataset %s' % name)
        return {'selected': name}

    def _select(self, name, validation):
        if validation:
            self.ctx.config.set('experiment/validation_dataset', name)
        else:
            self.ctx.config.set('dataset', name)

        self.ctx.config.write()

    def _get_ws(self, create_if_not_exist = False):
        if self.ws is None:
            self.ws = AzureProject(self.ctx)._get_ws(
                create_if_not_exist=create_if_not_exist)
        return self.ws

    def _columns(self, dataset = None, name = None):
        if not dataset:
            ws = self._get_ws()
            if name is None:
                name = self.ctx.config.get('dataset', None)
            if name is None:
                raise AzureException('Please specify dataset name...')
            dataset = Dataset.get_by_name(ws, name)

        df = dataset.take(1).to_pandas_dataframe()
        return df.columns.tolist()




def error_handler(decorated):
    def wrapper(self, *args, **kwargs):
        try:
            return decorated(self, *args, **kwargs)
        except Exception as exc:
            if self.ctx.debug:
                import traceback
                traceback.print_exc()

            if "ServiceException" in str(type(exc)):
                exc = Exception(exc.message)
                
            self.ctx.log(str(exc))
            if not hasattr(self.ctx, 'not_reraise_exceptions')\
               or not self.ctx.not_reraise_exceptions:
                raise exc
                
    return wrapper


def authenticated(decorated):
    def wrapper(self, *args, **kwargs):
        self.credentials.verify()
        return decorated(self, *args, **kwargs)
    return wrapper



import math

from HLL import HyperLogLog

class ProbabilisticCounter(object):
    # error_rate: 1% = 0.01, 0.5% = 0.005 (min 0.005)
    def __init__(self, error_rate=0.005):
        self.error_rate = error_rate

        # error_rate = 1.04 / sqrt(m)
        # m = 2 ** p -> registers count
        # M(1)... M(m) = 0 -> registers

        p = int(math.ceil(math.log((1.04 / error_rate) ** 2, 2)))
        self.hll = HyperLogLog(p)

    # returns: True - value is new, False, value already included
    def add(self, value):
        return self.hll.add(value)

    def count(self):
        return math.floor(self.hll.cardinality())

def get_p(error_rate):
    return int(math.ceil(math.log((1.04 / error_rate) ** 2, 2)))



import os
import json
import sys

from .impl.cloud.auth import AugerAuthApi
from a2ml.api.utils.base_credentials import BaseCredentials

class Credentials(BaseCredentials):
    """Manage credentials on user computer."""
    def __init__(self, ctx):
        super(Credentials, self).__init__(ctx, "auger")
        self.organization = None
        self.username = None
        self.api_url = None
        self.token = None

    def load(self):
        if hasattr(self.ctx, 'credentials') and self.ctx.credentials:
            content = self.ctx.credentials
        elif 'AUGER_CREDENTIALS' in os.environ:
            content = os.environ.get('AUGER_CREDENTIALS', None)
            content = json.loads(content) if content else {}
        else:
            self._ensure_credentials_file()
            with open(self.creds_file, 'r') as file:
                content = json.loads(file.read())

        self.username = content.get('username')
        self.organization = content.get('organization')
        if content.get('api_url'): 
            self.api_url = content['api_url']
        else:
            self.api_url = content.get('url', 'https://app.auger.ai')

        self.token = content.get('token')
            
        return self

    def login(self, username, password, organization, url=None):
        self.load()
        try:
            self.token = None
            self.save()

            if url is None:
                url = self.api_url

            token = AugerAuthApi(self.ctx).login(
                username, password, organization, url)

            self.token = token
            self.username = username
            self.api_url = url
            self.organization = organization
            self.save()

            self.ctx.log(
                'You are now logged in on %s as %s.' % (url, username))

        except Exception as exc:
            exc_text = str(exc)
            if 'Email or password incorrect' in exc_text:
                exc_text = 'Email or password incorrect...'
            self.ctx.log(exc_text)

    def logout(self):
        self.load()

        if self.token is None:
            self.ctx.log('You are not logged in Auger.')
        else:
            self.token = None
            self.api_url = None
            self.organization = None
            self.save()
            self.ctx.log('You are logged out of Auger.')

    def whoami(self):
        self.load()

        if self.token is None:
            self.ctx.log('Please login to Auger...')
        else:
            self.ctx.log(
                '%s %s %s' % (
                    self.username,
                    self.organization,
                    self.api_url))

    def serialize(self):
        return {
            'organization' : self.organization,
            'api_url': self.api_url,
            'token': self.token,
            'username': self.username
        }

    def save(self):
        self._ensure_credentials_file()

        content = self.serialize()
        with open(self.creds_file, 'w') as file:
            file.write(json.dumps(content))

    def verify(self):
        if self.token is None:
            raise Exception(
                'Please provide your credentials to Auger...')
        return True



import contextlib
import importlib
import os
import psutil
import sys
import threading
import time


from concurrent.futures import ThreadPoolExecutor, as_completed, thread

class ProviderRunner(object):
    """Runner executes provider jobs on threads."""
    class SpinningCursorThread(threading.Thread):
        def __init__(self,  *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)
            self._stop_event = threading.Event()

        def stop(self):
            self._stop_event.set()

        def stopped(self):
            return self._stop_event.is_set()

        def run(self):
            while not self.stopped():
                for cursor in '|/-\\':
                    sys.stdout.write(cursor)
                    sys.stdout.flush()
                    time.sleep(0.2)
                    sys.stdout.write('\b')

    def __init__(self, ctx, provider = None):
        super(ProviderRunner, self).__init__()
        self.ctx = ctx
        try:
            self.providers = self._load_providers(provider)
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise e

    def execute_one_provider(self, operation_name, *args, **kwargs):
        result = self.execute(operation_name, *args, **kwargs)

        if list(self.providers.keys())[0] in result:
            return list(result.values())[0]

        return result
            
    def execute(self, operation_name, *args, **kwargs):
        if len(self.providers) == 0:
            raise Exception('Please specify list of providers '
                'at config.yaml/providers')
        # if there is single operation requested
        # no need to run it on the thread
        if len(self.providers) == 1:
            provider_name = list(self.providers.keys())[0]
            return self.execute_provider(provider_name, operation_name, *args, **kwargs)

        with ThreadPoolExecutor(max_workers=len(self.providers)) as executor:
            futures = {
                executor.submit(
                    self.execute_provider, provider_name, operation_name, *args, **kwargs): provider_name
                for provider_name in self.providers.keys()
            }

            try:
                results = {}
                for future in as_completed(futures):
                    results.update(future.result())
                return results
            except KeyboardInterrupt:
                # not a graceful termination
                executor._threads.clear()
                thread._threads_queues.clear()
                raise

    def execute_provider(self, provider_name, operation_name, *args, **kwargs):
        try:
            provider = self.providers[provider_name]
            result = getattr(provider, operation_name)(*args, **kwargs)
            return {
                provider_name: {
                    'result': True,
                    'data': result
                }
            }
        except Exception as e:
            if self.ctx.debug:
                import traceback
                traceback.print_exc()

            return {
                provider_name: {
                    'result': False,
                    'data': str(e)
                }
            }

    def _load_providers(self, provider = None):
        def get_provider_class(p):
            module = importlib.import_module('a2ml.api.%s.a2ml' % p)
            return getattr(module, '%sA2ML' % p.capitalize())

        def get_instance(p):
            new_ctx = self.ctx.copy(p)
            
            if self.ctx.config.get('use_a2ml_hub', False) and p != 'auger':
                new_ctx.provider_info = {p: {"project": {"cluster": self.ctx.config.get("cluster", config_name=p)}}}
                provider_class = get_provider_class("auger")
            else:
                provider_class = get_provider_class(p)

            return provider_class(new_ctx)

        return {p: get_instance(p) for p in self.ctx.get_providers(provider)}



from a2ml.api.base_a2ml import BaseA2ML
from a2ml.api.utils.show_result import show_result

class A2MLModel(BaseA2ML):
    """Contains the model operations that interact with provider."""
    def __init__(self, ctx, provider=None):
        """Initializes a new a2ml model.

        Args:
            ctx (object): An instance of the a2ml Context.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider from methods.

        Returns:
            A2MLModel object

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx)
        """
        super(A2MLModel, self).__init__(ctx, "model")

        self.runner = self.build_runner(ctx, provider)
        self.local_runner = lambda: self.build_runner(ctx, provider, force_local=True)

    @show_result
    def deploy(self, model_id, locally=False, review=True, provider=None):
        """Deploy a model locally or to specified provider(s).

        Args:
            model_id (str): Model ID from the any experiment leaderboard.
            locally(bool): Deploys using a local model if True, on the Provider Cloud if False.
            review(bool): Should model support review based on actual data. The default is True.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': {'model_id': 'A017AC8EAD094FD'}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx).deploy(model_id='D881079E1ED14FB', locally=True)
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('deploy', model_id, locally, review)

    @show_result
    def predict(self, filename,
      model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None, provider=None):
        """Predict results with new data against deployed model. Predictions are stored next to the file with data to be predicted on. The file name will be appended with suffix _predicted.

        Args:
            filename(str): The file with data to request predictions for.
            model_id(str): The deployed model id you want to use.
            threshold(float): For classification models only. This will return class probabilities with response.
            locally(bool): Predicts using a local model if True, on the Provider Cloud if False.
            data: dict or array of records
            columns(list): list of column names if data is array of records
            predicted_at: Predict data date. Use for review of historical data.
            output(str): Output csv file path.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            if filename is not None. ::

                {
                    'result': True,
                    'data': {'predicted': 'dataset_predicted.csv'}
                }

            if filename is None and data is not None and columns is None. ::

                {
                    'result': True,
                    'data': {'predicted': [{col1: value1, col2: value2}, {col1: value3, col2: value4}]}
                }

            if filename is None and data is not None and columns is not None. ::

                {
                    'result': True,
                    'data': {'predicted': {'columns': ['col1', 'col2'], 'data': [['value1', 'value2'], ['value3', 'value4']]}}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                rv = A2MLModel(ctx).predict('../irises.csv', model_id)
                # if rv[provider].result is True
                # predictions are stored in rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [{'col1': 'value1', 'col2': 'value2'}, {'col1': 'value3', 'col2': 'value4'}]
                rv = A2MLModel(ctx).predict(None, model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [['value1', 'value2'], ['value3', 'value4']]
                columns = ['col1', 'col2']                
                rv = A2MLModel(ctx).predict(None, model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('predict', filename, model_id, threshold, locally, data, columns, predicted_at, output)

    @show_result
    def actuals(self, model_id, filename=None, actual_records=None, actuals_at=None, locally=False, provider=None):
        """Submits actual results(ground truths) for predictions of a deployed model. This is used to review and monitor active models.

        Note:
            It is assumed you have predictions against this model first. The file will need to fill in actual values for prediction_id.

            .. list-table:: actuals.csv
                :widths: 50 50
                :header-rows: 1

                * - prediction_id
                  - actual
                * - eaed9cd8-ba49-4c06-86d5-71d453c681d1
                  - Iris-setosa
                * - eaed9cd8-ba49-4c06-86d5-71d453c65290
                  - Iris-virginica

        Args:
            model_id(str): The deployed model id you want to use.
            filename(str): The file with data to request predictions for.
            actual_records: array of records [[prediction_id, actual]]
            actuals_at: Actuals date. Use for review of historical data.
            locally(bool): Process actuals locally.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': True
                }

            Errors. ::

                {
                    'auger': {
                        'result': False,
                        'data': 'Actual Prediction IDs not found in model predictions.'
                    }
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx).actuals('D881079E1ED14FB', filename=<path_to_file>/actuals.csv)

            .. code-block:: python
            
                # To pass just one actual:
                ctx = Context()
                actual_records = [['prediction_id', 'actual_value']]
                model = A2MLModel(ctx).actuals('D881079E1ED14FB', actual_records=actual_records)

            .. code-block:: python

                ctx = Context()
                actual_records = [['prediction_1', 'value1'], ['prediction_2', 'value2']]
                model = A2MLModel(ctx).actuals('D881079E1ED14FB', actual_records=actual_records)

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('actuals', model_id, filename, actual_records, actuals_at, locally)

    @show_result
    def review(self, model_id, locally=False, provider=None):
        """Review the performance of deployed model.

        Args:
            model_id(str): The deployed model id you want to use.
            locally(bool): Process review locally.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': {'score': {'accuracy': 0.8}}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx).review(model_id='D881079E1ED14FB')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('review', model_id)

    @show_result
    def undeploy(self, model_id, locally=False, provider=None):
        """Undeploy a model locally or from specified provider(s).

        Args:
            model_id (str): Model ID from any experiment leaderboard.
            locally(bool): Deploys using a local model if True, on the Provider Cloud if False. The default is False.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': {'model_id': 'A017AC8EAD094FD'}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx).undeploy(model_id='D881079E1ED14FB', locally=True)
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('undeploy', model_id, locally)



from .base import AugerBaseApi


class AugerActualApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api, prediction_id=None):
        super(AugerActualApi, self).__init__(
            ctx, pipeline_api, prediction_id)

    def create(self, records, actuals_at):
        params = {
            'pipeline_id': self.parent_api.object_id,
            'actuals': records
        }

        if actuals_at:
            params['actuals_at'] = str(actuals_at)

        return self._call_create(
            params=params, 
            has_return_object=False)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('deploy', short_help='Deploy trained model.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@click.argument('model-id', required=True, type=click.STRING)
@click.option('--locally', is_flag=True, default=False,
    help='Download and deploy trained model locally.')
@click.option('--review', is_flag=True, default=False,
    help='Should model support review based on actual data.')
@pass_context
def cmdl(ctx, provider, model_id, locally, review):
    """Deploy trained model."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).deploy(model_id, locally, review)



import click
from a2ml.api.a2ml_model import A2MLModel
from a2ml.api.utils.context import pass_context


@click.group('model', short_help='Model management')
@pass_context
def cmdl(ctx):
    """Model management"""
    ctx.setup_logger(format='')

@click.command('deploy', short_help='Deploy trained model.')
@click.argument('model-id', required=True, type=click.STRING)
@click.option('--locally', is_flag=True, default=False,
    help='Download and deploy trained model locally.')
@click.option('--review', is_flag=True, default=False,
    help='Should model support review based on actual data.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def deploy(ctx, provider, model_id, locally, review):
    """Deploy trained model."""
    A2MLModel(ctx, provider).deploy(model_id, locally, review)

@click.command('predict', short_help='Predict with deployed model.')
@click.argument('filename', required=True, type=click.STRING)
@click.option('--threshold', '-t', default=None, type=float,
    help='Threshold.')
@click.option('--model-id', '-m', type=click.STRING, required=True,
    help='Deployed model id.')
@click.option('--locally', is_flag=True, default=False,
    help='Predict locally using Docker image to run model.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@click.option('--output', '-o', type=click.STRING, required=False,
    help='Output csv file path.')
@pass_context
def predict(ctx, provider, filename, model_id, threshold, locally, output):
    """Predict with deployed model."""
    A2MLModel(ctx, provider).predict(filename, model_id, threshold=threshold, locally=locally, output=output)

@click.command('actuals', short_help='Send actual values for deployed model. Needed for review and monitoring.')
@click.argument('filename', required=True, type=click.STRING)
@click.option('--model-id', '-m', type=click.STRING, required=True,
    help='Deployed model id.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
help='Cloud AutoML Provider.')
@click.option('--locally', is_flag=True, default=False,
    help='Process actuals locally.')
@pass_context
def actuals(ctx, provider, filename, model_id, locally):
    """Predict with deployed model."""
    A2MLModel(ctx, provider).actuals(model_id, filename=filename, locally=locally)

@click.command('review', short_help='Review the performance of deployed model.')
@click.option('--model-id', '-m', type=click.STRING, required=True,
    help='Deployed model id.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def review(ctx, provider, model_id, output):
    """Predict with deployed model."""
    A2MLModel(ctx, provider).review(model_id)

@click.command('undeploy', short_help='Undeploy trained model.')
@click.argument('model-id', required=True, type=click.STRING)
@click.option('--locally', is_flag=True, default=False,
    help='Undeploy trained model locally.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def undeploy(ctx, provider, model_id, locally):
    """Undeploy trained model."""
    A2MLModel(ctx, provider).undeploy(model_id, locally)

@pass_context
def add_commands(ctx):
    cmdl.add_command(deploy)
    cmdl.add_command(predict)
    cmdl.add_command(actuals)
    cmdl.add_command(review)
    cmdl.add_command(undeploy)

add_commands()



import os
import csv

from ..cloud.pipeline import AugerPipelineApi
from a2ml.api.utils import fsclient

class ModelActual():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelActual, self).__init__()
        self.ctx = ctx

    def execute(self, model_id, filename, actual_records, actuals_at):
        if filename:
            self.ctx.log('Sending actuals on data in %s' % filename)
            if filename and not fsclient.is_s3_path(filename):
                filename = os.path.abspath(filename)

            return self._file_actual_to_cloud(filename, model_id, actuals_at)
        else:
            return self._actuals_to_cloud(actual_records, model_id, actuals_at)

    def _file_actual_to_cloud(self, filename, model_id, actuals_at):
        with open(filename) as f:
          a = [{k: v for k, v in row.items()}
              for row in csv.DictReader(f, skipinitialspace=True)]

        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        return pipeline_api.actual(a, actuals_at)

    def _actuals_to_cloud(self, actual_records, model_id, actuals_at):
        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        return pipeline_api.actual(actual_records, actuals_at)



import boto3
import math
import os
import sys
import threading

from boto3.s3.transfer import TransferConfig
from urllib.parse import urlparse
from uuid import uuid4

from  a2ml.api.utils import fsclient

class FileUploader(object):
    def __init__(self, bucket, endpoint_url, access_key, secret_key, session_token):
        self.bucket = bucket

        self.client = boto3.client(
            's3',
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=session_token,
        )

    def transfer_config(self):
        return TransferConfig(
            multipart_threshold=1024 * 25,
            max_concurrency=4,
            multipart_chunksize=1024 * 25,
            use_threads=True
        )

    def multi_part_upload(self, file_path, key=None, callback=None):
        if not key:
            file_name = os.path.basename(file_path)
            base_name, extension = file_name.split('.', 1)
            key = 'uploads/' + base_name + '-' + str(uuid4())[0:8] + '.' + extension

        self.client.upload_file(
            file_path,
            self.bucket,
            key,
            Config=self.transfer_config(),
            Callback=callback
        )

        return f's3://{self.bucket}/{key}'

    def multipart_upload_obj(self, file_obj, key, callback=None):
        self.client.upload_fileobj(
            file_obj,
            self.bucket,
            key,
            Config=self.transfer_config(),
            Callback=callback
        )

        return 's3://%s/%s' % (self.bucket, key)

    def download(self, s3_path, local_path):
        url = urlparse(s3_path)
        self.client.download_file(self.bucket, url.path, local_path)

class OnelineProgressPercentage(object):
    def __init__(self, filename):
        self._filename = os.path.basename(filename)
        self._size = float(fsclient.get_file_size(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()
        self._done = False

    def __call__(self, bytes_amount):
        with self._lock:
            if not self._done:
                self._seen_so_far += bytes_amount
                percentage = (self._seen_so_far / self._size) * 100
                sys.stdout.write(
                    "\r%s  %s / %s  (%.2f%%)" % (self._filename, self._seen_so_far, int(self._size), percentage)
                )
                sys.stdout.flush()

                if percentage >= 100:
                    self._done = True
                    sys.stdout.write("\n")

class NewlineProgressPercentage(object):
    def __init__(self, filename, report_step = 25):
        self._filename = os.path.basename(filename)
        self._size = float(fsclient.get_file_size(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()
        self._reported_steps = 0
        self._report_step = report_step

    def __call__(self, bytes_amount):
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            if (math.floor(percentage / self._report_step) > self._reported_steps):
                self._reported_steps = math.floor(percentage / self._report_step)

                sys.stdout.write(
                    "%s  %s / %s  (%.2f%%)\n" % (self._filename, self._seen_so_far, int(self._size), percentage)
                )
                sys.stdout.flush()



from functools import wraps
import sys

from .exceptions import (AugerException)

def _get_project(self, autocreate):
    from .project import Project

    project_name = self.ctx.config.get('name', None)
    if project_name is None:
        raise AugerException(
            'Please specify project name in auger.yaml/name...')
    project = Project(self.ctx, project_name)
    project_properties = project.properties()
    if project_properties is None:
        if autocreate:
            self.ctx.log(
                'Can\'t find project %s on the Auger Cloud. '
                'Creating...' % project_name)
            project.create()
        else:
            raise AugerException('Can\'t find project %s' % project_name)
    return project

def with_project(autocreate=False):
    def decorator(decorated):
        @wraps(decorated)
        def wrapper(self, *args, **kwargs):
            project = _get_project(self, autocreate)
            return decorated(self, project, *args, **kwargs)
        return wrapper
    return decorator

def with_dataset(decorated):
    from .dataset import DataSet

    def wrapper(self, *args, **kwargs):
        project = _get_project(self, False)
        data_set_name = self.ctx.config.get('dataset', None)
        if data_set_name is None:
            raise AugerException(
                'Please specify dataset name in auger.yaml/dataset...')
        dataset = DataSet(self.ctx, project, data_set_name)
        return decorated(self, dataset, *args, **kwargs)
    return wrapper



import time
from .base import AugerBaseApi
from .cluster import AugerClusterApi
from ..exceptions import AugerException


class AugerProjectApi(AugerBaseApi):
    """Auger Project API."""

    def __init__(self, ctx, org_api,
        project_name=None, project_id=None):
        super(AugerProjectApi, self).__init__(
            ctx, org_api, project_name, project_id)
        assert org_api is not None, 'Organization must be set for Project'
        self._set_api_request_path('AugerProjectApi')

    def is_running(self):
        return self.properties().get('status') == 'running'

    def create(self):
        return self._call_create({
            'name': self.object_name, 'organization_id': self.parent_api.oid})

    def start(self):
        self._ensure_object_id()
        project_properties = self.properties()

        status = project_properties.get('status')
        if status == 'running':
            return project_properties
        project_status = ['deployed', 'deploying']
        if status in project_status:
            return self.wait_for_status(project_status)

        cluster_id = project_properties.get('cluster_id')
        cluster_api = AugerClusterApi(self.ctx, self, cluster_id)

        if self.parent_api.get_cluster_mode() == 'single_tenant':
            if not cluster_api.is_running():
                cluster_api.create()
        else:
            cluster_settings = cluster_api.get_cluster_settings(self.ctx)
            # self.rest_api.call('update_project', {
            #     'id': self.object_id,
            #     'cluster_autoterminate_minutes':
            #         cluster_settings.get('autoterminate_minutes')})
            try:
                self.rest_api.call('deploy_project', {
                    'id': self.object_id,
                    'worker_type_id': cluster_settings.get('worker_type_id'),
                    'workers_count' : cluster_settings.get('workers_count'),
                    'kubernetes_stack': cluster_settings.get('kubernetes_stack')})
            except:
                project_properties = self.properties()
                status = project_properties.get('status')
                project_status = ['deployed', 'deploying', 'running']
                if not status in project_status:
                    raise                    

        result = self.wait_for_status(
            ['undeployed', 'deployed', 'scaling', 'zero_scaled', 'deploying'])
        time.sleep(20)
        return result

    def stop(self):
        if self.status() != 'undeployed':
            self.rest_api.call(
                'undeploy_project', {'id': self.object_id})
            return self.wait_for_status(['running', 'undeploying'])




from celery import Celery
from celery.concurrency import asynpool
from a2ml.api.utils.json_utils import convert_simple_numpy_type, convert_nan_inf
from a2ml.tasks_queue.config import Config
from kombu import Exchange, Queue
from kombu.serialization import register
from kombu.utils import json as _json
import os
import warnings

task_config = Config()


# Disable warnings
def warn(*args, **kwargs):
    pass


warnings.warn = warn

celeryApp = Celery(
    'a2ml.tasks_queue',
    include=[
        'a2ml.tasks_queue.tasks_api',
        'a2ml.tasks_queue.tasks_hub_api'
    ]
)

celeryApp.conf.enable_utc = True

celeryApp.conf.broker_url = task_config.broker_url or "redis://%s:%s/0" % (
    os.environ.get('REDIS_HOST', 'localhost'),
    os.environ.get('REDIS_PORT', '6379')
)

# TODO: for AMQP there better to use "rpc://"
# see https://docs.celeryproject.org/en/4.0/whatsnew-4.0.html#features-removed-for-lack-of-funding
# now leave it optional and not RPC by default
if celeryApp.conf.broker_url.lower().startswith('amqp'):
    if os.environ.get('RESULT_RPC', 'false').lower() == 'true':
        celeryApp.conf.result_backend = "rpc://"
    else:
        celeryApp.conf.result_backend = "amqp://"
else:
    celeryApp.conf.result_backend = celeryApp.conf.broker_url

# Next settings for result queues should be absolutely
# the same as declared in consumer else
# amqp.exceptions.PreconditionFailed: Queue.declare: (406)
# PRECONDITION_FAILED will be raised
celeryApp.conf.result_expires = int(
    os.environ.get('CELERY_RESULT_EXPIRES_SEC', 86400)
)
celeryApp.conf.result_persistent = True
celeryApp.conf.task_acks_late = False
celeryApp.conf.task_send_sent_event = True

celeryApp.conf.task_queues = (
    Queue('a2ml', Exchange('a2ml', delivery_mode=1),
          routing_key='a2ml', durable=False, priority=3),
)

celeryApp.conf.task_routes = {
    'a2ml.tasks_queue.tasks_hub_api.*': {'queue': 'a2ml'},
}

class NumpyKombuJSONEncoder(_json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        res = convert_simple_numpy_type(obj)
        if res is not None:
            return res

        return super(NumpyKombuJSONEncoder, self).default(obj)


# Encoder function
def my_dumps(obj):
    obj = convert_nan_inf(obj)
    return _json.dumps(obj, cls=NumpyKombuJSONEncoder)

register(
    'myjson', my_dumps, _json.loads,
    content_type='application/json',
    content_encoding='utf-8'
)

# Tell celery to use your new serializer:
# celeryApp.conf.accept_content = ['myjson']
celeryApp.conf.task_serializer = 'myjson'
celeryApp.conf.result_serializer = 'myjson'

# try monkey patch startup timeout since
# we take longer than 4.0 seconds to startup
asynpool.PROC_ALIVE_TIMEOUT = 60.0



import click
import os

from a2ml.api.utils.context import pass_context

@click.command('worker', short_help='Run Celery worker')
@click.option('--pool', '-p', type=click.STRING, required=False, default='gevent',
    show_default=True, help='host interface to bind server')
@click.option('--concurrency', '-c', type=click.INT, required=False, default=100,
    show_default=True, help='port to bind server')
@pass_context
def cmdl(ctx, pool, concurrency):
    cmd_line = f'celery -A a2ml.tasks_queue.celery_app worker --loglevel=info --pool={pool} -c {concurrency}'
    os.system(cmd_line)
